;;;;
;;;; e n v o b j . s c m	     -- reified environments
;;;;

(provide 'envobj)
(require 'lib)
(require 'vectree)

"
Environments:

A binding is an association pair indicating
that a symbol is bound to some value.
cons cells are not used, for the reasons listed in the 'vectree module,
as well as because this allows the same infrastructure
to be later used for proof environments,
where instead of symbol->value binding
(understood as the axiom ``this symbol, as an expression,
reduces to this form''), we can have arbitrary axioms.
A binding (bound x t) is then considered as the axiom
(both% x (T<% x t))

A frame is a list of bindings

An environment is a list of frames

"/;"

;;;
;;; Bindings
;;;

;;; Basic constructors and destruct(urat)ors
(define binding-tag (make-tag 'bound))
(define (make-binding sym val) (vector binding-tag sym val))
(define binding? (vectree-detector binding-tag 2))
(define binding:sym v-1st)
(define binding:val v-2nd)
(define binding:set-sym! v-set-1st!)
(define binding:set-val! v-set-2nd!)

(define (binding:matchsym sym)
  (lambda (binding)
    (eq? (binding:sym binding) sym)))

;;;
;;; Frames
;;;

;;; Basic constructors and destruct(urat)ors
(define frame-tag (make-tag 'frame))
(define (make-frame bindings) (vector frame-tag bindings))
(define (frame:new) (make-frame '()))

(define frame:bindings v-1st)
(define frame:set-bindings! v-set-1st!)
(define (frame:add-binding! frame binding)
  (frame:set-bindings! frame (cons binding (frame:bindings frame))))

(define (frame:copy frame)
  (make-frame (frame:bindings frame)))

(define (frame:cons binding frame)
  (make-frame (cons binding (frame:bindings frame))))
(define (frame:car binding frame)
  (car (frame:bindings frame)))
(define (frame:cdr binding frame)
  (make-frame (cdr (frame:bindings frame))))

(define (frame:lookup frame fun found not-found)
  (list:lookup (frame:bindings frame) fun found not-found))
(define (frame:find frame sym found not-found)
  (frame:lookup
    frame
    (binding:matchsym sym)
    (lambda (binding t) (found binding))
    not-found))
(define (frame:try-get frame sym found not-found)
  (frame:find frame sym
      (lambda (binding) (found (binding:val binding)))
      not-found))
(define (frame:get frame sym)
  (frame:try-get frame sym
      (lambda (val) val)
      (lambda () (soft-error "not found symbol" sym "in frame" frame))))



(define (frame:fill-binding! frame sym val)
  (frame:lookup
     frame
     (binding:matchsym sym)
     (lambda (binding t)
       (eq? (binding:val binding) val))
     (lambda ()
       (frame:add-binding! frame (make-binding sym val)))))


;;;
;;; Environments
;;;

;;; Basic constructors and destruct(urat)ors
(define env-tag
  (make-tag 'env))
(define (make-env frames)
  (vector env-tag frames))
(define (frame->env frame)
  (make-env (list frame)))
(define (env:new)
  (frame->env (frame:new)))
(define env? (vectree-detector env-tag 1))

(define env:frames
  v-1st)
(define env:set-frames!
  v-set-1st!)
(define (env:add-frame! env frame)
  (env:set-frames! env (cons frame (env:frames env))))
(define (env:add-new-frame! env frame)
  (env:add-frame! env (frame:new)))
(define (env:cdr! env)
  (env:set-frames! env (cdr (env:frames env))))
(define (env:copy env)
  (make-env (env:frames env)))

(define (env:cons frame env)
  (make-env (cons frame (env:frames env))))
(define (env:car env)
  (car (env:frames env)))
(define (env:cdr frame env)
  (make-env (cdr (env:frames env))))
(define (env:cons-new-frame env frame)
  (env:cons (frame:new) env))

(define (env:add-binding! env b)
  (frame:add-binding! (env:car env) b))
(define (env:add-definition! env sym val)
  (env:add-binding! env (make-binding sym val)))

(define (env:cons-binding env b)
  (let* ((frames (env:frames env))
	 (frame1 (car frames))
	 (restframes (cdr frames))
	 (newframe1 (frame:cons b frame1))
	 (newframes (cons frame1 restframes)))
  (make-env (newframes))))
(define (env:cons-definition env sym val)
  (env:cons-binding env (make-binding sym val)))

(define (env:direct-lookup env fun found not-found)
  (list:lookup (env:frames env) fun found not-found))
(define env:lookup-bindings
  (indirect-lookup env:direct-lookup frame:lookup))
(define (env:find env sym found not-found)
  (env:lookup-bindings
    env
    (binding:matchsym sym)
    (lambda (binding t) (found binding))
    not-found))
(define (env:try-get env sym found not-found)
  (env:find env sym
      (lambda (binding) (found (binding:val binding)))
      not-found))
(define (env:bound? env sym)
  (env:find env sym
      (lambda (binding) #t)
      (lambda () #f)))
(define (env:get env sym)
  (env:try-get env sym
      (lambda (val) val)
      (lambda () (soft-error "not found symbol" sym "in env" env))))

(define (env:lookup-with-depth env fun found not-found)
  (let again
    ((frames (env:frames env))
     (n 0))
    (if (null? frames)
	(not-found n)
	(frame:lookup (car frames) (lambda (x) (fun x))
	   (lambda (b r) (found b n r))
	   (lambda () (again (cdr frames) (+ n 1)))))))
(define (env:find-with-depth env sym found not-found)
  (env:lookup-with-depth
    env
    (binding:matchsym sym)
    (lambda (b n t) (found b n))
    not-found))
(define (env:try-get-with-depth env sym found not-found)
  (env:find-with-depth env sym
      (lambda (b n) (found (binding:val b) n))
      not-found))
(define (env:get-depth env sym)
  (env:try-get env sym
      (lambda (val n) n)
      (lambda (n) (soft-error "not found symbol" sym "in env" env))))

(define (env:empty? env)
  (lookup->empty? env:lookup-bindings))
(define (env:merge . envs)
  (make-env (apply append (map env:frames envs))))
