;;;;
;;;; v e c t r e e . s c m	     -- trees of tagged vectors,
;;;;					as used to represent expressions
;;;;					and various structures.
;;;;

;(#include "stdheader.h")

(provide 'vectree)
(require 'lib)
(require 'tags)


"\
This module provides infrastructure for implementing
abstract data types in Scheme (something that lacks too much).
   I feel that representing everything with lists and cons cells is
too low-level and too confusing. I higher-level program specification style
that uses well differentiated constructors leads to
(1) easier typing, which allows for better ``optimization'' by
  an efficient compiler, including global removal of tag-checking,
  and even of redundant tags altogether, eliminating space and time
  overhead from naive implementations;
(2) easier typing, which eases the port to languages with
  a rich static type system (like the ML and Haskell families).
(3) easier typing, which prevents bugs from human confusion
  when manipulating nested structures.
(4) an efficient implementation should resolve all tagging overhead
   problems nicely
"

;; All objects are meant to be read-only, but for explicit references
;;
;; We trust the program not to modify anything but this emulated type...


(define (v-tag x)	(vector-ref x 0))
(define (v-1st x)	(vector-ref x 1))
(define (v-2nd x)	(vector-ref x 2))
(define (v-3rd x)	(vector-ref x 3))
(define (v-4th x)	(vector-ref x 4))
(define (v-5th x)	(vector-ref x 5))

(define (v-set-tag! x v)	(vector-set! x 0 v))
(define (v-set-1st! x v)	(vector-set! x 1 v))
(define (v-set-2nd! x v)	(vector-set! x 2 v))
(define (v-set-3rd! x v)	(vector-set! x 3 v))
(define (v-set-4th! x v)	(vector-set! x 4 v))
(define (v-set-5th! x v)	(vector-set! x 5 v))

(define (v-field-getter n)
  (lambda (x)
    (vector-ref x n)))

(define (v-field-setter n)
  (lambda (x v)
    (vector-set! x n v)))

(define sys$$vector? vector?)
(define (vector-l>=n? x n)
  (and (vector? x) (>= (vector-length x) n)))
(define (vector-l>n? x n)
  (and (vector? x) (> (vector-length x) n)))
(define (vectree? obj)
  (and (vector-l>n? obj 0) (tags$$is-tag? (v-tag obj))))

(define (vectree-maker tag)
  (lambda args
    (apply vector tag args)))

(define (vectree-detector tag minlen)
  (lambda (obj)
    (and (vector-l>n? obj minlen) (eq? (v-tag obj) tag))))
